import datetime
import json
from collections import defaultdict
from typing import Dict, List, Optional, Set

from pendulum import DateTime, UTC
from pytz.tzinfo import BaseTzInfo
from sqlalchemy.orm import Session as SASession

from duty_board.models.calendar import Calendar
from duty_board.models.on_call_event import OnCallEvent
from duty_board.models.person import Person
from duty_board.plugin.helpers.duty_calendar_config import DutyCalendarConfig
from duty_board.web_helpers.response_types import _Calendar, _Events, _ExtraInfoOnPerson, _Person, PersonResponse


def format_datetime_for_timezone(dt: datetime.datetime, timezone: BaseTzInfo) -> str:
    dt_tz_aware = dt.astimezone(timezone)
    return dt_tz_aware.strftime("%Y-%m-%d %H:%M:%S")


def _get_events_ending_from_now_onwards(
    session: SASession, all_encountered_person_uids: Set[int], timezone: BaseTzInfo
) -> Dict[str, List[_Events]]:
    result = session.query(OnCallEvent).filter(OnCallEvent.end_event_utc >= DateTime.utcnow()).all()
    mapped: Dict[str, List[_Events]] = defaultdict(list)
    for calendar_event in result:
        mapped[calendar_event.calendar_uid].append(
            _Events(
                start_event=format_datetime_for_timezone(calendar_event.start_event_utc, timezone),
                end_event=format_datetime_for_timezone(calendar_event.end_event_utc, timezone),
                person_uid=calendar_event.person_uid,
            )
        )
        all_encountered_person_uids.add(calendar_event.person_uid)
    return mapped


def get_calendars(session: SASession, all_encountered_person_uids: Set[int], timezone: BaseTzInfo) -> List[_Calendar]:
    events: Dict[str, List[_Events]] = _get_events_ending_from_now_onwards(
        session=session, all_encountered_person_uids=all_encountered_person_uids, timezone=timezone
    )
    result = session.query(Calendar).order_by(Calendar.order).all()
    return [
        _Calendar(
            uid=single_calendar.uid,
            name=single_calendar.name,
            description=single_calendar.description,
            category=single_calendar.category,
            order=single_calendar.order,
            last_update=format_datetime_for_timezone(single_calendar.last_update_utc, timezone),
            error_msg=single_calendar.error_msg or "",
            sync=single_calendar.sync,
            events=events.get(single_calendar.uid) or [],
        )
        for single_calendar in result
    ]


def get_persons(session: SASession, all_person_uids: Set[int], timezone: BaseTzInfo) -> Dict[int, _Person]:
    result = session.query(Person).where(Person.uid.in_(all_person_uids))
    return {
        a_person.uid: _Person(
            uid=a_person.uid,
            username=a_person.username,
            email=a_person.email,
            img_filename=a_person.img_filename,
            extra_attributes={},
            last_update=format_datetime_for_timezone(a_person.last_update_utc, timezone),
            error_msg=a_person.error_msg or "",
            sync=a_person.sync,
        )
        for a_person in result
    }


def parse_extra_attributes(person_uid: int, extra_attributes_str: Optional[str]) -> List[_ExtraInfoOnPerson]:
    if extra_attributes_str is None or len(extra_attributes_str) == 0:
        return []

    extra_attributes_list = json.loads(extra_attributes_str)
    if not isinstance(extra_attributes_list, dict):
        raise ValueError(f"Expected extra_attributes field to be a dict for {person_uid=}.")

    result_list: List[_ExtraInfoOnPerson] = []
    value: Dict[str, str]
    for key, value in extra_attributes_list.items():
        if not isinstance(value, dict):
            raise ValueError(f"Expected extra_attributes[{key}] field to be a dict for {person_uid=}.")
        if "information" not in value.keys():
            raise ValueError(f"Missing required extra_attributes[{key}][information] for {person_uid=}.")
        result_list.append(
            _ExtraInfoOnPerson(
                information=value.get("information"),
                icon=value.get("icon", "FaMinus"),
                icon_color=value.get("icon_color", "black"),
                url=value.get("url", None),
            )
        )
    return result_list


def get_person(session: SASession, person_uid: int, timezone: BaseTzInfo) -> PersonResponse:
    person = session.query(Person).where(Person.uid == person_uid).first()
    if person is None:
        raise ValueError(f"Invalid {person_uid=} passed.")

    return PersonResponse(
        uid=person.uid,
        username=person.username,
        email=person.email,
        img_filename=person.img_filename,
        extra_attributes=parse_extra_attributes(person_uid, extra_attributes_str=person.extra_attributes_json),
        last_update=format_datetime_for_timezone(person.last_update_utc, timezone),
        error_msg=person.error_msg or "",
        sync=person.sync,
    )


def _create_or_update_calendar(session: SASession, calendar: DutyCalendarConfig) -> None:
    calendar_db_instance = session.query(Calendar).where(Calendar.uid == calendar.uid).first()
    if calendar_db_instance is not None:
        calendar_db_instance.name = calendar.name
        calendar_db_instance.description = calendar.description
        calendar_db_instance.category = calendar.category
        calendar_db_instance.order = calendar.order
        calendar_db_instance.icalendar_url = calendar.icalendar_url
        calendar_db_instance.event_prefix = calendar.event_prefix
        session.merge(calendar_db_instance)
    else:
        calendar_db_instance = Calendar(
            uid=calendar.uid,
            name=calendar.name,
            description=calendar.description,
            category=calendar.category,
            order=calendar.order,
            icalendar_url=calendar.icalendar_url,
            event_prefix=calendar.event_prefix,
            error_msg=None,
            last_update_utc=DateTime(1970, 1, 1, 0, 0, 0, tzinfo=UTC),
            sync=True,
        )
        session.add(calendar_db_instance)


def sync_duty_calendar_configurations_to_postgres(
    session: SASession, duty_calendar_configurations: List[DutyCalendarConfig]
) -> None:
    for duty_calendar_config in duty_calendar_configurations:
        _create_or_update_calendar(session, duty_calendar_config)
    all_described_calendar_uids: List[str] = [dcc.uid for dcc in duty_calendar_configurations]
    session.query(Calendar).where(Calendar.uid.not_in(all_described_calendar_uids)).delete()



import logging
import os
import sys
from pathlib import Path
from typing import Dict, Final, List, Set

import pytz
from fastapi import FastAPI, HTTPException
from pytz.exceptions import UnknownTimeZoneError
from pytz.tzinfo import BaseTzInfo
from sqladmin import Admin
from starlette.middleware.cors import CORSMiddleware
from starlette.responses import FileResponse, HTMLResponse
from starlette.staticfiles import StaticFiles
from tzlocal import get_localzone

from duty_board.alchemy import add_sqladmin, queries, settings
from duty_board.alchemy.session import create_session
from duty_board.models import generate_fake_data
from duty_board.plugin.abstract_plugin import AbstractPlugin
from duty_board.plugin.helpers import plugin_fetcher
from duty_board.web_helpers.gzip_static_files import GZIPStaticFiles
from duty_board.web_helpers.response_types import _Calendar, _Config, _Person, CurrentSchedule, PersonResponse

logging.basicConfig(
    stream=sys.stdout,
    level=logging.DEBUG,
    format="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
    datefmt="%m-%d %H:%M",
)
logger = logging.getLogger(__name__)

app = FastAPI()
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_methods=["*"],
    allow_headers=["*"],
)

plugin: AbstractPlugin
admin: Admin

CURRENT_DIR: Final[Path] = Path(__file__).absolute().parent
print(f"{CURRENT_DIR=}")
if settings.SQL_ALCHEMY_CONN:
    settings.configure_orm()
    plugin = plugin_fetcher.get_plugin()
    app.mount("/dist", GZIPStaticFiles(directory=CURRENT_DIR / "www" / "dist", check_dir=False), name="dist")
    app.mount("/static", StaticFiles(directory=CURRENT_DIR / "www" / "static"), name="static")
    app.mount("/person_img", StaticFiles(directory=plugin.absolute_path_to_user_images_folder), name="person_img")
    admin = add_sqladmin.add_sqladmin(app=app, plugin=plugin)
if os.environ.get("CREATE_DUMMY_RECORDS", "") == "1":
    generate_fake_data.create_fake_database_rows_if_not_present()


def _parse_timezone_str(timezone_str: str) -> BaseTzInfo:
    try:
        return pytz.timezone(timezone_str)
    except UnknownTimeZoneError:
        return get_localzone()


def _get_config_object(timezone_object: BaseTzInfo) -> _Config:
    return _Config(
        text_color=plugin.text_color_hex,
        background_color=plugin.background_color_hex,
        categories=plugin.category_order,
        git_repository_url=plugin.git_repository_url,
        enable_admin_button=plugin.enable_admin_button,
        timezone=timezone_object.zone,
    )


@app.get("/get_schedule", response_model=CurrentSchedule)
async def get_schedule(timezone: str):
    timezone_object = _parse_timezone_str(timezone)
    config = _get_config_object(timezone_object)
    with create_session() as session:
        all_encountered_person_uids: Set[int] = set()
        calendars: List[_Calendar] = queries.get_calendars(
            session=session, all_encountered_person_uids=all_encountered_person_uids, timezone=timezone_object
        )
        persons: Dict[int, _Person] = queries.get_persons(
            session=session, all_person_uids=all_encountered_person_uids, timezone=timezone_object
        )
        return CurrentSchedule(config=config, calendars=calendars, persons=persons)


@app.get("/get_person", response_model=PersonResponse)
async def get_person(person_uid: int, timezone: str):
    timezone_object = _parse_timezone_str(timezone)
    with create_session() as session:
        return queries.get_person(session=session, person_uid=person_uid, timezone=timezone_object)


@app.get(
    "/company_logo.png",
    response_description="Returns a thumbnail image from a larger image",
    responses={200: {"description": "Company logo", "content": {"image/png": {}}}},
)
def thumbnail_image():
    file_path = plugin.absolute_path_to_company_logo_png
    if os.path.exists(file_path):
        return FileResponse(file_path, media_type="image/jpeg", filename="company_logo.png")
    raise HTTPException(status_code=500, detail="Company logo could not be found.")


@app.get("/{full_path:path}", response_class=HTMLResponse, include_in_schema=False)
async def accept_all():
    return FileResponse(CURRENT_DIR / "www" / "dist" / "index.html")


# @app.get("/duty", response_class=HTMLResponse)
# async def get_schedule():
#     return FileResponse("duty_board/www/dist/index.html")



from __future__ import annotations

import datetime
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Optional

from sqlalchemy.orm import Session as SASession

from duty_board.models.calendar import Calendar
from duty_board.models.person import Person
from duty_board.plugin.helpers.duty_calendar_config import DutyCalendarConfig


class AbstractPlugin(ABC):
    admin_session_length: datetime.timedelta = datetime.timedelta(days=7)
    person_update_frequency: datetime.timedelta = datetime.timedelta(hours=1)
    calendar_update_frequency: datetime.timedelta = datetime.timedelta(days=1)
    background_color_hex: str = "#3C9C2D"
    text_color_hex: str = "white"
    absolute_path_to_company_logo_png: Path | None = None
    # Path on the back-end to the user images folder.
    absolute_path_to_user_images_folder: Path | None = None
    category_order: List[str] = []
    duty_calendar_configurations: List[DutyCalendarConfig] = []
    enable_admin_button: bool = True
    git_repository_url: Optional[str] = "https://github.com/Jorricks/DutyBoard"

    def __init__(self, *args, **kwargs):
        super(AbstractPlugin, self).__init__(*args, **kwargs)

    @abstractmethod
    def sync_person(self, person: Person, session: SASession) -> Person:
        pass

    @abstractmethod
    def sync_calendar(self, calendar: Calendar, session: SASession) -> Calendar:
        pass

    @abstractmethod
    async def admin_login_attempt(self, username: str, password: str) -> bool:
        pass



from sqlalchemy import Boolean, Column, Index, Integer, String
from sqlalchemy.orm import relationship

from duty_board.alchemy.settings import Base
from duty_board.alchemy.sqlalchemy_types import UtcDateTime


class Calendar(Base):
    __table_args__ = (Index("calendar_last_update_utc", "last_update_utc"),)
    __tablename__ = "calendar"
    uid = Column(String(50), primary_key=True, nullable=False)
    name = Column(String(200), nullable=False)
    description = Column(String(5000), nullable=True)
    category = Column(String(50), nullable=False)
    order = Column(Integer, nullable=False)
    icalendar_url = Column(String(500), nullable=True)
    event_prefix = Column(String(50), nullable=True)
    error_msg = Column(String(9999), nullable=True, comment="If any, the error of the latest sync attempt.")
    last_update_utc = Column(UtcDateTime(), nullable=False)
    sync = Column(Boolean(), default=True, nullable=False)
    events = relationship("OnCallEvent", back_populates="calendar", cascade="all, delete-orphan")

    def __repr__(self):
        return f"Calendar(uid={self.uid})"



from sqlalchemy import Boolean, Column, Index, Integer, String

from duty_board.alchemy.settings import Base
from duty_board.alchemy.sqlalchemy_types import UtcDateTime


class Person(Base):
    __table_args__ = (Index("person_last_update_utc", "last_update_utc"),)
    __tablename__ = "person"
    uid = Column(Integer, primary_key=True, autoincrement=True, unique=True)
    username = Column(String(50), nullable=True, unique=True)
    email = Column(String(50), nullable=True, unique=True)
    img_filename = Column(String(100), nullable=True)
    extra_attributes_json = Column(String(100000), nullable=True, comment="Extra attributes represented as a json.")
    error_msg = Column(String(9999), nullable=True, comment="If any, the error of the latest sync attempt.")
    last_update_utc = Column(UtcDateTime(), nullable=False)
    sync = Column(Boolean(), default=True, nullable=False)

    def __repr__(self) -> str:
        return f"Person(username='{self.username}', email='{self.email}')"



import logging
import time
import traceback
from typing import Optional

from pendulum import DateTime
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import Query
from sqlalchemy.orm import Session as SASession

from duty_board.alchemy import queries, settings
from duty_board.alchemy.session import create_session
from duty_board.models.calendar import Calendar
from duty_board.models.on_call_event import OnCallEvent
from duty_board.models.person import Person
from duty_board.plugin.abstract_plugin import AbstractPlugin
from duty_board.plugin.helpers import plugin_fetcher

logger = logging.getLogger(__name__)


def get_most_outdated_person(plugin: AbstractPlugin, session: SASession) -> Optional[Person]:
    update_persons_with_last_update_before: DateTime = DateTime.utcnow() - plugin.person_update_frequency
    return (
        session.query(Person)
        .where(Person.last_update_utc <= update_persons_with_last_update_before)
        .order_by(Person.last_update_utc)
        .first()
    )


def get_most_outdated_calendar(plugin: AbstractPlugin, session: SASession) -> Optional[Calendar]:
    update_calendars_with_last_update_before: DateTime = DateTime.utcnow() - plugin.calendar_update_frequency
    return (
        session.query(Calendar)
        .where(Calendar.last_update_utc <= update_calendars_with_last_update_before)
        .order_by(Calendar.last_update_utc)
        .first()
    )


def ensure_person_uniqueness(new_person: Person, session: SASession) -> Person:
    """
    This function ensures we don't have 1000 users with the except same username and or email.
    If there is a new person object with the same username and email, we wipe the old person object and replace the
    UIDS.
    """
    query: Query = (
        session.query(Person)
        .filter(Person.uid != new_person.uid)
        .filter(Person.username == new_person.username)
        .filter(Person.email == new_person.email)
    )
    for same_person in query.all():
        logger.warning(f"Deleting {same_person=} in favor of {new_person=}.")
        session.delete(same_person)
        new_values = {"person_uid": new_person.uid}
        session.query(OnCallEvent).filter(OnCallEvent.person_uid == same_person.uid).update(new_values)
        logger.info(f"Also updated all references of {same_person.uid=} to {new_person.uid=} in OnCallEvents table.")
    return new_person


def update_all_outdated_persons(plugin: AbstractPlugin) -> None:
    person: Optional[Person] = None
    while True:
        with create_session() as session:
            try:
                person = get_most_outdated_person(plugin=plugin, session=session)
                if person is None:
                    logger.debug("Nothing to update here :).")
                    break
                logger.info(f"Updating {person}.")
                try:
                    person = plugin.sync_person(person=person, session=session)
                    logger.debug("Successfully executed plugins sync_person().")
                    person = ensure_person_uniqueness(new_person=person, session=session)
                    person.error_msg = None
                except Exception:
                    logger.exception(f"Failed to update {person}.")
                    person.error_msg = traceback.format_exc()
                finally:
                    person.last_update_utc = DateTime.utcnow()
                    session.merge(person)
            except Exception:
                logger.exception(f"Failed to update {person} in the database? Maybe there is some database error?")
        logger.info("Successfully finished updating person.")


def update_all_outdated_calendars(plugin: AbstractPlugin) -> None:
    calendar: Optional[Calendar] = None
    for _ in range(3):  # Update at most x calendars before we resort to updating persons.
        with create_session() as session:
            try:
                calendar = get_most_outdated_calendar(plugin=plugin, session=session)
                if calendar is None:
                    logger.debug("Nothing to update here :).")
                    break
                logger.info(f"Updating {calendar}.")
                try:
                    calendar = plugin.sync_calendar(calendar=calendar, session=session)
                    logger.debug("Successfully executed plugins sync_calendar().")
                    calendar.error_msg = None
                except Exception:
                    logger.exception(f"Failed to update {calendar}.")
                    calendar.error_msg = traceback.format_exc()
                finally:
                    calendar.last_update_utc = DateTime.utcnow()
                    session.merge(calendar)
                logger.debug(f"Successfully updated {calendar}.")
            except Exception:
                logger.exception(f"Failed to update {calendar} in the database? Maybe there is some database error?")
        logger.info("Successfully finished updating calendar.")


def enter_loop():
    settings.configure_orm()
    plugin: AbstractPlugin = plugin_fetcher.get_plugin()

    logger.info("Updating plugin calendars in the database.")
    with create_session() as session:
        queries.sync_duty_calendar_configurations_to_postgres(session, plugin.duty_calendar_configurations)

    while True:
        try:
            logger.info("Starting another update pass.")
            update_all_outdated_calendars(plugin=plugin)
            time.sleep(1)
            update_all_outdated_persons(plugin=plugin)
            time.sleep(1)
        except SQLAlchemyError:
            logger.exception("Error occurred when updating the database.")
            time.sleep(1)



from typing import Any, Dict, List, Optional

from pydantic import BaseModel


class _Person(BaseModel):
    uid: int
    username: Optional[str]
    email: Optional[str]
    img_filename: Optional[str]
    extra_attributes: Dict[str, Any]
    last_update: str
    error_msg: str
    sync: bool


class _Events(BaseModel):
    start_event: str
    end_event: str
    person_uid: int


class _Calendar(BaseModel):
    uid: str
    name: str
    description: str
    category: str
    order: int
    last_update: str
    error_msg: str
    sync: bool
    events: List[_Events]


class _Config(BaseModel):
    timezone: str
    text_color: str
    background_color: str
    categories: List[str]
    git_repository_url: Optional[str]
    enable_admin_button: bool


class CurrentSchedule(BaseModel):
    config: _Config
    calendars: List[_Calendar]
    persons: Dict[int, _Person]


class _ExtraInfoOnPerson(BaseModel):
    information: str
    icon: str
    icon_color: str
    url: Optional[str]


class PersonResponse(BaseModel):
    uid: str
    username: Optional[str]
    email: Optional[str]
    img_filename: Optional[str]
    extra_attributes: List[_ExtraInfoOnPerson]
    last_update: str
    error_msg: str
    sync: bool



from pathlib import Path

from sqlalchemy.orm import Session as SASession

from duty_board.models.person import Person
from duty_board.plugin.abstract_plugin import AbstractPlugin
from duty_board.plugin.ical_plugin_mixin import ICalPluginMixin


class ExamplePlugin(ICalPluginMixin, AbstractPlugin):
    absolute_path_to_company_logo_png = Path(__file__).resolve().parent / "example_logo.png"
    absolute_path_to_user_images_folder = Path(__file__).resolve().parent.parent.parent / "www" / "user_images"
    category_order = ["Big Data", "Infrastructure"]

    def sync_person(self, person: Person, session: SASession) -> Person:
        """This is to be implemented by our end user."""
        return person

    # sync_calendar is implemented by the ICalPluginMixin.

    async def admin_login_attempt(self, username: str, password: str) -> bool:
        return username == "admin" and password == "admin123"
